#include <opencv2/opencv.hpp>

#include <fstream>

void loadLandmarks(std::string path, std::vector<cv::Point2f> &points) 
{
    std::ifstream fs(path);
    for (;;) {
        int i, x, y;
        if (fs >> i >> x >> y) points.push_back(cv::Point(x, y));
        else break;
    }
    fs.close();
}

static int u = 0;
void displayPoints(cv::Mat &image, std::vector<cv::Point2f> &points)
{
    u++;
    cv::Mat img = image.clone();
    for (auto point : points) {
        cv::circle(img,point,2*img.rows/300, cv::Scalar(0,0,255),cv::FILLED);
    }
    cv::Mat tmp;
    cv::resize(img, tmp, cv::Size(img.cols * 300 / img.rows, 300));
    cv::imshow("landmarks"+std::to_string(u), tmp);
    cv::waitKey(0);
}

std::vector<cv::Point2f> forehead (std::vector<cv::Point2f> &points)
{
    std::vector<int> brows {25, 23, 20, 18};
    std::vector<int> eyes {45, 47, 40, 36};
    std::vector<cv::Point2f> addons;
    for (size_t i=0; i<4; i++) addons.push_back(points[brows[i]]+2*(points[brows[i]]-points[eyes[i]]));
    return addons;
}

void displayTriangulation(cv::Mat &image, std::vector<cv::Point2f> &points, std::vector< std::vector<int> > &triangulation)
{
    u++;
    cv::Mat img = image.clone();
    for (auto &indices : triangulation) {
        std::vector<cv::Point2f> p;
        for (int index : indices) p.push_back(points[index]);
        cv::line(img, p[0], p[1], cv::Scalar(0, 0, 255), 2 * img.rows / 300);
        cv::line(img, p[1], p[2], cv::Scalar(0, 0, 255), 2 * img.rows / 300);
        cv::line(img, p[2], p[0], cv::Scalar(0, 0, 255), 2 * img.rows / 300);
    }
    cv::Mat tmp;
    cv::resize(img, tmp, cv::Size(img.cols * 300 / img.rows, 300));
    cv::imshow("triangulation"+std::to_string(u), tmp);
    cv::waitKey(0);
}

void warpTriangle (cv::Mat &image1, cv::Mat &image2, std::vector<cv::Point2f> &triangle1, std::vector<cv::Point2f> &triangle2)
{
    cv::Rect rect1 = cv::boundingRect(triangle1);
    cv::Rect rect2 = cv::boundingRect(triangle2);
    cv::Point2f offset1 = rect1.tl();
    cv::Point2f offset2 = rect2.tl();
    std::vector<cv::Point2f> triangle1_, triangle2_;
    for (auto point : triangle1) triangle1_.push_back(point-offset1);
    for (auto point : triangle2) triangle2_.push_back(point-offset2);
    cv::Mat subimage1 = image1(rect1);
    cv::Mat subimage2 = image2(rect2);
    cv::Mat transform = cv::getAffineTransform(triangle1_, triangle2_);
    cv::Mat warped;
    cv::warpAffine(subimage1,warped,transform,subimage2.size(),cv::INTER_LINEAR,cv::BORDER_REFLECT_101);
    cv::Mat mask2(subimage2.size(),CV_8UC1,cv::Scalar(0));
    std::vector<cv::Point> polygon;
    for (auto point : triangle2_) polygon.push_back(cv::Point(int(point.x),int(point.y)));
    cv::fillPoly(mask2, polygon, cv::Scalar(255));
    warped.copyTo(subimage2,mask2);
}

int main() {
    
    // Read images
    cv::Mat img1 = cv::imread("../Wrinkles.jpg");
    cv::Mat img2 = cv::imread("../Face.jpg");

    // Load landmarks
    std::vector<cv::Point2f> points1, points2;
    loadLandmarks("../Wrinkles.txt",points1);
    loadLandmarks("../Face.txt",points2);
    displayPoints(img1, points1);
    displayPoints(img2, points2);

    // Add forehead points
    for (auto point : forehead(points1)) points1.push_back(point);
    for (auto point : forehead(points2)) points2.push_back(point);
    displayPoints(img1, points1);
    displayPoints(img2, points2);   

    // Find triangulation
    cv::Rect rect(0,0,img2.cols,img2.rows);
    cv::Subdiv2D delaunay(rect);
    delaunay.insert(points2);
    std::vector<cv::Vec6f> triangleList;
    delaunay.getTriangleList(triangleList);
    std::vector< std::vector<cv::Point2f> > triangles;
    for (cv::Vec6f &t : triangleList) {
        triangles.push_back(std::vector<cv::Point2f> {cv::Point2f(t[0],t[1]),cv::Point2f(t[2],t[3]),cv::Point2f(t[4],t[5])});
    }
    auto localizer = [](std::vector<cv::Point2f> vec, cv::Point2f item) { return std::find(vec.begin(), vec.end(), item) - vec.begin(); };
    std::vector< std::vector<int> > triangulation;
    for (auto &triangle : triangles) {
        std::vector<int> indices;
        for (auto point : triangle) indices.push_back(int(localizer(points2,point)));
        triangulation.push_back(indices);
    }
    displayTriangulation(img2,points2,triangulation);
    displayTriangulation(img1,points1,triangulation);

    // Morphing: Warp img1 onto img2
    cv::Mat img1Warped = img2.clone();
    for (std::vector<int> &indices : triangulation) {
        std::vector<cv::Point2f> triangle1, triangle2;
        for (int index : indices) {
            triangle1.push_back(points1[index]);
            triangle2.push_back(points2[index]);
        }
        warpTriangle(img1,img1Warped,triangle1,triangle2);
    }
    displayTriangulation(img1Warped,points2,triangulation);
    cv::imshow("warped",img1Warped);
    cv::waitKey(0);
    
    // Mixed seemless cloning
    cv::Mat mask(img2.size(),CV_8UC1,cv::Scalar(0));
    std::vector<cv::Point2f> hull;
    cv::convexHull(points2, hull);
    std::vector<cv::Point> polygon;
    for (auto point : hull) polygon.push_back(cv::Point(int(point.x), int(point.y)));
    cv::fillPoly(mask, polygon, cv::Scalar(255));
    cv::imshow("mask",mask);
    cv::waitKey(0);
    cv::Rect r = cv::boundingRect(hull);
    cv::Point2f pivot = (r.tl()+r.br()) / 2.0;
    cv::Mat cloned;
    cv::seamlessClone(img1Warped, img2, mask, pivot, cloned, cv::MIXED_CLONE);
    cv::imshow("cloned",cloned);
    cv::waitKey(0);

    cv::destroyAllWindows();
    
    return 0;
}